/*
 *  CPRNG.c
 *
 *  Author  : Ahmed Obied (ahmed.obied@gmail.com)
 *  Version : 0.1
 *
 *  CPRNG implements the Blum-Blum-Shub pseudorandom number generator.    
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include "CPRNG.h"

// if defined then seed from /dev/random else seed from /dev/urandom
//#define	__SECURE_RANDOM_DEVICE__	
#define	ENTROPY_BYTES			80
#define FACTOR_BYTES			64

// seed the Blum-Blum-Shub pseudorandom number generator by 
// generating random p, q and s 
CPRNG::CPRNG()
{
	int fd, bytes_num;
        char *data, *s_data;
        mpz_t p, q, seed,tmp;
	
	// start seeding
	mpz_init(p);
	mpz_init(q);
	mpz_init(seed);
	mpz_init(tmp);
	mpz_init(n);
	mpz_init(p_x);
	mpz_init(c_x);
        data   = (char *)malloc(FACTOR_BYTES);
	s_data = (char *)malloc(ENTROPY_BYTES);
	memset(data,0,FACTOR_BYTES);
	memset(s_data,0,ENTROPY_BYTES);
	// use the secure device /dev/random for the entropy pool
#ifdef __SECURE_RANDOM_DEVICE__
        fd = open("/dev/random",O_RDONLY);
        if (fd < 0) {
                perror("/dev/random");
                exit(-1);
        }
#else
	// use /dev/urandom for faster and less secure entropy pool
	fd = open("/dev/urandom",O_RDONLY);
	if (fd < 0) {
		perror("/dev/urandom");
		exit(-1);
	}
#endif
	// generate a random p
	while(1) {
		// read the random bytes for p (64 bytes)
		while ((bytes_num = read(fd,data,FACTOR_BYTES)) != FACTOR_BYTES);
		if (bytes_num < 0) {
			perror("CPRNG: read for p");
			exit(-1);
		}
		mpz_import(p,1,1,FACTOR_BYTES,0,0,data);
		// p must be a prime 
		mpz_nextprime(p,p);
		// p must be congruent to 3 mod 4
		if (mpz_congruent_ui_p(p,3,4))
			break;
	}
	// generate a random q
	while(1) {	
		// read the random bytes for q (64 bytes)
		while ((bytes_num = read(fd,data,FACTOR_BYTES)) != FACTOR_BYTES);
		if (bytes_num < 0) {
			perror("CPRNG: read for q");
			exit(-1);
		}
        	mpz_import(q,1,1,FACTOR_BYTES,0,0,data);
		// q must be a prime
		mpz_nextprime(q,q);
		// q must be congruent to 3 mod 4
		if (mpz_congruent_ui_p(q,3,4))
			break;
	}
	// n = p * q (n is 128 bytes = 1024-bits)
        mpz_mul(n,p,q);
	// generate the seed -> s
	while(1) {
   		// read the random bytes for s (80-bytes) and iterate until we read the exact
		// number
		while((bytes_num = read(fd,s_data,ENTROPY_BYTES)) != ENTROPY_BYTES);
		if (bytes_num < 0) {
			perror("CPRNG: read for seed");
			exit(-1);
		}
		mpz_import(seed,1,1,ENTROPY_BYTES,0,0,s_data);
		// gcd(s,n) must be 1 else we re-generate
		mpz_gcd(tmp,n,seed);
                if (!mpz_cmp_ui(tmp,1))
                        break; 	
       	}
	// compute x_0 = s^{2} mod n 
	mpz_powm_ui(p_x,seed,2,n);
	close(fd);
        free(data);
	free(s_data);
	mpz_clear(p);
	mpz_clear(q);
	mpz_clear(seed);
	mpz_clear(tmp);
}

CPRNG::~CPRNG()
{
	mpz_clear(p_x);
	mpz_clear(c_x);
	mpz_clear(n);
}

// implement the Blum-Blum-Shub algorithm
void CPRNG::generate_random_number(int size, mpz_t &prand)
{
	int i;
	
	// set pseudorandom number to zero first
	mpz_set_ui(prand,0);
	// iterate by taking the previous value to the power 2 mod n
        for (i = 0 ; i < size ; i++) {
               	mpz_powm_ui(c_x,p_x,2,n);
		
		// if the least-significant bit of the ith number is set then
		// then set the ith bit in the generated pseudorandom number
               	if (mpz_tstbit(c_x,0))
                       	mpz_setbit(prand,i);
		// previous x becomes current x and so on
		mpz_set(p_x,c_x);
       	}
}


